
/*
*****************************************************************************
*
*                          Copyright (c) 2000 2005
*
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice, 
*     this list of conditions and the following disclaimer.
*  - Redistributions in binary form must reproduce the above copyright notice, 
*     this list of conditions and the following disclaimer in the documentation
*     and/or other materials provided with the distribution.
*  - The names of its contributors may not be used to endorse or promote 
*     products derived from this software without specific prior written 
*     permission.
*   
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
*  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
*  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
*  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
*  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*  POSSIBILITY OF SUCH DAMAGE.
*
******************************************************************************
* File          : cryptmod.c
* Author        : Zadig
* Creation date : 27/01/2005
*.............................................................................
* Description :
******************************************************************************
*/

/* ########################################################################## */
/* ### #include                                                               */
#include <drivers/Drivers.h>
#include <drivers/KernelExport.h>
#include "CRYPT.h"

/* ########################################################################## */
/* ### Functions                                                              */
static status_t std_ops(int32 op, ...);
int MOD_cryptCreateContext( C_OUT CRYPT_CONTEXT C_PTR cryptContext,
      C_IN CRYPT_USER cryptUser, C_IN CRYPT_ALGO_TYPE cryptAlgo );
int MOD_cryptDestroyContext( C_IN CRYPT_CONTEXT cryptContext );

int MOD_cryptEncrypt( C_IN CRYPT_CONTEXT cryptContext, 
                        C_INOUT void C_PTR buffer, C_IN int length );
int MOD_cryptDecrypt( C_IN CRYPT_CONTEXT cryptContext, 
                        C_INOUT void C_PTR buffer, C_IN int length );

int MOD_cryptSetAttribute( C_IN CRYPT_HANDLE cryptHandle,
      C_IN CRYPT_ATTRIBUTE_TYPE attributeType, C_IN int value );
int MOD_cryptSetAttributeString( C_IN CRYPT_HANDLE cryptHandle,
      C_IN CRYPT_ATTRIBUTE_TYPE attributeType,
      C_IN void C_PTR value, C_IN int valueLength );
int MOD_cryptGetAttribute( C_IN CRYPT_HANDLE cryptHandle,
      C_IN CRYPT_ATTRIBUTE_TYPE attributeType,
      C_OUT int C_PTR value );
int MOD_cryptGetAttributeString( C_IN CRYPT_HANDLE cryptHandle,
      C_IN CRYPT_ATTRIBUTE_TYPE attributeType,
      C_OUT void C_PTR value,
      C_OUT int C_PTR valueLength );
int MOD_cryptDeleteAttribute( C_IN CRYPT_HANDLE cryptHandle,
      C_IN CRYPT_ATTRIBUTE_TYPE attributeType );


/* ########################################################################## */
/* ### Variables                                                              */
static struct cryptlib_module_info  gs_module_info =
{
   {
      CRYPTLIB_MODULE_NAME,
      B_KEEP_LOADED,
      std_ops
   },
   MOD_cryptCreateContext,
   MOD_cryptDestroyContext,
   MOD_cryptEncrypt,
   MOD_cryptDecrypt,
   MOD_cryptSetAttribute,
   MOD_cryptSetAttributeString,
   MOD_cryptGetAttribute,
   MOD_cryptGetAttributeString,
   MOD_cryptDeleteAttribute
};

_EXPORT module_info *modules[] =
{
   (module_info*) &gs_module_info,
   NULL
};

static status_t std_ops(int32 op, ...)
{
   switch(op)
   {
      case B_MODULE_INIT:
         if(cryptInit() != CRYPT_OK)
         {
            dprintf("cryptmod: unable to init cryptlib\n");
            return(B_ERROR);
         }
         break;

      case B_MODULE_UNINIT:
         cryptEnd();
         break;

      default:
         dprintf("cryptmod: unknown op (0x%X)\n", (unsigned int)op);
         return(B_ERROR);
   }
   return(B_OK);
}


int MOD_cryptCreateContext( C_OUT CRYPT_CONTEXT C_PTR cryptContext,
      C_IN CRYPT_USER cryptUser, C_IN CRYPT_ALGO_TYPE cryptAlgo )
{
   int i_ret;
   
   i_ret = cryptCreateContext(cryptContext, cryptUser, cryptAlgo );

   return(i_ret);
}

int MOD_cryptDestroyContext( C_IN CRYPT_CONTEXT cryptContext )
{
   int i_ret;
   
   i_ret = cryptDestroyContext(cryptContext );

   return(i_ret);
}

int MOD_cryptEncrypt( C_IN CRYPT_CONTEXT cryptContext, 
                        C_INOUT void C_PTR buffer, C_IN int length )
{
   int i_ret;
   
   i_ret = cryptEncrypt( cryptContext, buffer, length );

   return(i_ret);
}

int MOD_cryptDecrypt( C_IN CRYPT_CONTEXT cryptContext, 
                        C_INOUT void C_PTR buffer, C_IN int length )
{
   int i_ret;
   
   i_ret = cryptDecrypt( cryptContext, buffer, length );

   return(i_ret);
}

int MOD_cryptSetAttribute( C_IN CRYPT_HANDLE cryptHandle,
      C_IN CRYPT_ATTRIBUTE_TYPE attributeType, C_IN int value )
{
   int i_ret;
   
   i_ret = cryptSetAttribute( cryptHandle, attributeType, value );

   return(i_ret);
}

int MOD_cryptSetAttributeString( C_IN CRYPT_HANDLE cryptHandle,
      C_IN CRYPT_ATTRIBUTE_TYPE attributeType,
      C_IN void C_PTR value, C_IN int valueLength )
{
   int i_ret;
   
   i_ret = cryptSetAttributeString( cryptHandle, attributeType, 
                                 value, valueLength );

   return(i_ret);
}

int MOD_cryptGetAttribute( C_IN CRYPT_HANDLE cryptHandle,
      C_IN CRYPT_ATTRIBUTE_TYPE attributeType,
      C_OUT int C_PTR value )
{
   int i_ret;
   
   i_ret = cryptGetAttribute( cryptHandle, attributeType, value );

   return(i_ret);
}

int MOD_cryptGetAttributeString( C_IN CRYPT_HANDLE cryptHandle,
      C_IN CRYPT_ATTRIBUTE_TYPE attributeType,
      C_OUT void C_PTR value,
      C_OUT int C_PTR valueLength )
{
   int i_ret;
   
   i_ret = cryptGetAttributeString( cryptHandle, attributeType,
                                    value, valueLength );

   return(i_ret);
}

int MOD_cryptDeleteAttribute( C_IN CRYPT_HANDLE cryptHandle,
      C_IN CRYPT_ATTRIBUTE_TYPE attributeType )
{
   int i_ret;
   
   i_ret = cryptDeleteAttribute( cryptHandle, attributeType );

   return(i_ret);
}

