#include <stdio.h>
#include <String.h>

#include "LoopPoolMessage.h"

LoopPoolMessage::LoopPoolMessage(char *psz_File) : BMessage('Pool')
{
   pc_ConfigFile = new BFile(psz_File, B_READ_WRITE | B_CREATE_FILE);
   Unflatten(pc_ConfigFile);
   pc_ConfigFile->Seek(0,SEEK_SET);
}

LoopPoolMessage::~LoopPoolMessage()
{
   delete pc_ConfigFile;
}

status_t LoopPoolMessage::Flush(void)
{   
   uint16   i_Version;
   ssize_t  i_Size;
   status_t i_ret;
   
   if(FindInt16((const char*)LOOP_VERSION_FIELD, (int16*)&i_Version) != B_OK)
      AddInt16(LOOP_VERSION_FIELD, LOOP_VERSION);
   
   i_Size = FlattenedSize();
   i_ret = Flatten(pc_ConfigFile, &i_Size);
   pc_ConfigFile->Seek(0,SEEK_SET);
   return(i_ret);
}

status_t LoopPoolMessage::GetFirstLoop(LoopMessage *pc_Loop)
{
   i_LoopIndex = 0;
   return(GetNextLoop(pc_Loop));
}

status_t LoopPoolMessage::GetNextLoop(LoopMessage *pc_Loop)
{
   pc_Loop->MakeEmpty();
   if(FindMessage(LOOP_LOOP_FIELD, i_LoopIndex, pc_Loop) == B_OK)
   {
      i_LoopIndex++;
      return(B_OK);
   }

   return(B_ENTRY_NOT_FOUND);
}

status_t LoopPoolMessage::SetLoop(LoopMessage *pc_Loop)
{
   LoopMessage c_LoopMessage;
   BString c_Name1;
      
   if(pc_Loop->FindString(LOOP_NAME_FIELD, &c_Name1) != B_OK)
      return(B_ERROR);

   if(FindLoop((char*)c_Name1.String(), &c_LoopMessage) == B_OK)
   {
      DelLoop(pc_Loop);
      return(AddLoop(pc_Loop));
   }

   return(B_ERROR);
}

status_t LoopPoolMessage::AddLoop(LoopMessage *pc_Loop)
{
   int32 i_CurIndex = 0;
   LoopMessage c_LoopMessage;
   BString c_Name1;
   BString c_Name2;
      
   if(pc_Loop->FindString(LOOP_NAME_FIELD, &c_Name1) != B_OK)
      return(B_ERROR);
   
   /* check that the loop is not already in the message */
   while(FindMessage(LOOP_LOOP_FIELD, i_CurIndex, &c_LoopMessage) == B_OK)
   {
      if(c_LoopMessage.FindString(LOOP_NAME_FIELD, &c_Name2) == B_OK)
      {
         if(c_Name1 == c_Name2) /* already in pool */
            return(B_ERROR);
      }
      
      i_CurIndex++;
   }
        
   return(AddMessage(LOOP_LOOP_FIELD, pc_Loop));
}

status_t LoopPoolMessage::DelLoop(LoopMessage *pc_Loop)
{
   uint8 i_CurIndex = 0;
   LoopMessage c_LoopMessage;
   BString  c_Name1, c_Name2;
   
   if(pc_Loop->FindString(LOOP_NAME_FIELD, &c_Name1) != B_OK)
      return(B_ERROR);
   
   /* find the loop is in the message */
   while(FindMessage(LOOP_LOOP_FIELD, i_CurIndex, &c_LoopMessage) == B_OK)
   {
      if(pc_Loop->FindString(LOOP_NAME_FIELD, &c_Name2) == B_OK)
      {
         if(c_Name1 == c_Name2) 
            return(RemoveData(LOOP_LOOP_FIELD, i_CurIndex));  /* and delete */
      }
      i_CurIndex++;
   }
      
   return(B_ENTRY_NOT_FOUND);
}



status_t LoopPoolMessage::FindLoop(char *sz_Name, LoopMessage *pc_Loop)
{
   BString     c_LoopName;
   
   if(GetFirstLoop(pc_Loop) == B_OK)
   {
      pc_Loop->GetName(&c_LoopName);
      if(strcmp(sz_Name, c_LoopName.String()) == 0)
         return(B_OK);
      
      while(GetNextLoop(pc_Loop) == B_OK) 
      {          
         pc_Loop->GetName(&c_LoopName);
         if(strcmp(sz_Name, c_LoopName.String()) == 0)
            return(B_OK);
      }
   }
   return(B_ERROR);
}


