
/*
*****************************************************************************
*
*                          Copyright (c) 2000 2005
*
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice, 
*     this list of conditions and the following disclaimer.
*  - Redistributions in binary form must reproduce the above copyright notice, 
*     this list of conditions and the following disclaimer in the documentation
*     and/or other materials provided with the distribution.
*  - The names of its contributors may not be used to endorse or promote 
*     products derived from this software without specific prior written 
*     permission.
*   
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
*  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
*  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
*  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
*  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*  POSSIBILITY OF SUCH DAMAGE.
*
******************************************************************************
* File          : PrefWindow.cpp
* Author        : Zadig
* Creation date : 23/08/2005
*.............................................................................
* Description :
******************************************************************************
*/

/* ########################################################################## */
/* ### #include                                                               */
#include <stdio.h>

#include <Application.h>
#include <Deskbar.h>
#include <Roster.h>

#include "PrefMessages.h"
#include "PrefWindow.h"
#include "SettingsWindow.h"
#include "TextAlert.h"

#include "RepView.h"

/* ########################################################################## */
/* ### Variables                                                              */

/* ########################################################################## */
/* ### Functions                                                              */

PrefWindow::PrefWindow(LoopPoolMessage *pc_LoopPool)  :
   BWindow(BRect(0, 0, W_WIDTH, W_HEIGHT), 
            "Edit Loops", B_TITLED_WINDOW_LOOK, 
            B_NORMAL_WINDOW_FEEL, 
            B_ASYNCHRONOUS_CONTROLS | B_NOT_RESIZABLE | B_NOT_ZOOMABLE )
{
   float i_Width, i_Height;
   BRect   c_ButtonView;
   status_t i_ret;
   LoopMessage c_Loop;
   BString     c_LoopName;
   BDeskbar deskbar;
 
   if(pc_LoopPool == NULL)
   {
      /* load settings */
      BDirectory  c_ConfigDir;
      c_ConfigDir.CreateDirectory(LOOP_CONFIG_PATH, NULL);
      pc_LoopPoolMessage = new LoopPoolMessage(LOOP_CONFIG_FILE);
      if(pc_LoopPoolMessage == NULL)
         return;
   }
   else
   {
      pc_LoopPoolMessage = pc_LoopPool;
   }

   /* set window position */
   BScreen  c_Screen;
   BRect c_WindowPosition(0, 0, W_WIDTH, W_HEIGHT);

   c_WindowPosition = c_Screen.Frame();
   c_WindowPosition.left = c_WindowPosition.right - c_WindowPosition.left;
   c_WindowPosition.left /= 2;
   c_WindowPosition.left -= (W_WIDTH/2);
   c_WindowPosition.right = c_WindowPosition.left + W_WIDTH;

   c_WindowPosition.top = c_WindowPosition.bottom - c_WindowPosition.top;
   c_WindowPosition.top /= 2;
   c_WindowPosition.top -= (W_HEIGHT/2);
   c_WindowPosition.bottom = c_WindowPosition.top + W_HEIGHT;

   MoveTo(c_WindowPosition.left, c_WindowPosition.top);
 
   /* build gui */
   pc_MainView = new BView(Bounds(), "MainView",          
         B_FOLLOW_LEFT,B_NAVIGABLE);
   pc_MainView->SetViewColor(BG_COLOR);
   AddChild(pc_MainView);

   pc_LoopView = new LoopView(pc_LoopPoolMessage, pc_MainView, Bounds().top);
   i_ret = pc_LoopPoolMessage->GetFirstLoop(&c_Loop);
   while(i_ret == B_OK)
   {
      if(c_Loop.GetName(&c_LoopName) == B_OK)
         pc_LoopView->AddLoop((char*)c_LoopName.String());

      i_ret = pc_LoopPoolMessage->GetNextLoop(&c_Loop);
   }

   /* apply/cancel/deskbar buttons */
   c_ButtonView = Bounds();
   c_ButtonView.top += 170;
   c_ButtonView.left += 15;

   pc_ApplyButton = new BButton(c_ButtonView,"Apply Changes", "Apply",
         new BMessage(PREF_MSG_APPLY_LOOP));
   pc_ApplyButton->GetPreferredSize(&i_Width,&i_Height);
   pc_ApplyButton->ResizeTo(i_Width,i_Height);
   pc_MainView->AddChild(pc_ApplyButton);

   pc_CancelButton = new BButton(c_ButtonView,"Cancel Changes", "Cancel",
         new BMessage(PREF_MSG_QUIT));
   pc_CancelButton->GetPreferredSize(&i_Width,&i_Height);
   pc_CancelButton->ResizeTo(i_Width,i_Height);
   pc_CancelButton->MoveTo(pc_ApplyButton->Frame().right + 20,c_ButtonView.top);
   pc_MainView->AddChild(pc_CancelButton);

   SetDefaultButton(pc_CancelButton);

   Show();
}

PrefWindow::~PrefWindow()
{
}

void PrefWindow::MessageReceived(BMessage *message)
{
   TextAlert      *pc_TextAlert;
   SettingsWindow *pc_SettingsWindow;
   BString        c_LoopName;
   LoopMessage    *pc_NewLoop, c_TempLoop;
   BDeskbar       deskbar;
   RepView        *pc_RepView;
   
   switch (message->what)
   {
      case PREF_MSG_ADD_LOOP:
         pc_TextAlert = new TextAlert(false, "Enter loop name");
         if(pc_TextAlert != NULL)
         {
            c_LoopName = pc_TextAlert->Go();
            if(pc_LoopPoolMessage->FindLoop((char*)c_LoopName.String(), &c_TempLoop) != B_OK)
            {
               pc_SettingsWindow = new SettingsWindow(pc_LoopPoolMessage,
                     c_LoopName.String());
               if(pc_SettingsWindow != NULL)
               {
                  Lock();
                  pc_NewLoop = pc_SettingsWindow->Go();
                  if(pc_NewLoop != NULL)
                  {
                     pc_LoopView->AddLoop((char*)c_LoopName.String());
                  }
                  Unlock();
               }
            }
            else
            {
            }
         }
         break;

      case PREF_MSG_DEL_LOOP:
         pc_LoopView->DelActiveLoop();
         break;

      case PREF_MSG_APPLY_LOOP:
         pc_LoopPoolMessage->Flush();
         break;
      
      case PREF_MSG_QUIT:
         Lock();
         Quit();
         Unlock();
         break;

      case LOOP_MSG_INVOKE:
         if(pc_LoopView->GetActiveLoopName(&c_LoopName) == B_OK)
         {
            pc_SettingsWindow = new SettingsWindow(pc_LoopPoolMessage,
                  c_LoopName.String());

            if(pc_SettingsWindow != NULL)
               pc_SettingsWindow->Go();
         }
         break;

      default:
         BWindow::MessageReceived(message);
         break;
   }
}

bool PrefWindow::QuitRequested()
{
   return true;
}


