
/*
*****************************************************************************
*
*                          Copyright (c) 2000 2005
*
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice, 
*     this list of conditions and the following disclaimer.
*  - Redistributions in binary form must reproduce the above copyright notice, 
*     this list of conditions and the following disclaimer in the documentation
*     and/or other materials provided with the distribution.
*  - The names of its contributors may not be used to endorse or promote 
*     products derived from this software without specific prior written 
*     permission.
*   
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
*  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
*  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
*  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
*  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*  POSSIBILITY OF SUCH DAMAGE.
*
******************************************************************************
* File          : SettingsWindow.cpp
* Author        : Zadig
* Creation date : 28/08/2005
*.............................................................................
* Description :
******************************************************************************
*/

/* ########################################################################## */
/* ### #include                                                               */
#include <stdlib.h>
#include <stdio.h>
#include <ClassInfo.h>

#include "SettingsWindow.h"

/* ########################################################################## */
/* ### Variables                                                              */

/* ########################################################################## */
/* ### Functions                                                              */

SettingsWindow::SettingsWindow(LoopPoolMessage *pc_LoopPoolMessage, const char *sz_Title)  :
   BWindow(BRect(0, 0, SETTINGS_WINDOW_WIDTH, SETTINGS_WINDOW_HEIGHT), 
            sz_Title, B_TITLED_WINDOW_LOOK, 
            B_MODAL_APP_WINDOW_FEEL, 
            B_ASYNCHRONOUS_CONTROLS | B_NOT_RESIZABLE | B_NOT_ZOOMABLE )
{
   uint16   i_Number;
   uint32   i_Algo;
   float    i_Width, i_Height;
   BRect    c_ButtonView, c_ViewRect;
   BPoint   c_Point;
   BString  c_LoopName, c_LoopPath;

   /* set loop message settings */
   pc_LoopMessage = new LoopMessage();
   pc_LoopPool = pc_LoopPoolMessage;

   if(pc_LoopPoolMessage->FindLoop((char*)sz_Title, pc_LoopMessage) != B_OK)
   {
      /* loop does not exists yet */
      c_LoopName.SetTo(sz_Title);
      pc_LoopMessage = new LoopMessage();
      pc_LoopMessage->SetName(&c_LoopName);
      pc_LoopMessage->SetNumber(0);
      pc_LoopMessage->SetAlgo(CRYPT_ALGO_AES);
      pc_LoopMessage->SetKeyType(0);
      b_NewLoop = true;
   }
   else
   {
      b_NewLoop = false;
   }
   b_Cancel = false;

   /* build gui */
   pc_MainView = new BView(Bounds(), "MainView",          
         B_FOLLOW_LEFT,B_NAVIGABLE);
   pc_MainView->SetViewColor(SETTINGS_WINDOW_BG_COLOR);
   AddChild(pc_MainView);

   /* settings box */
   c_ViewRect = pc_MainView->Bounds();
   c_ViewRect.left += 5;
   c_ViewRect.right -= 5;
   c_ViewRect.top += 10;
   c_ViewRect.bottom = c_ViewRect.top + 80;

   pc_SettingsBox = new BBox(c_ViewRect);
   pc_SettingsBox->SetLabel("Settings");
   pc_MainView->AddChild(pc_SettingsBox);

   c_ButtonView = pc_SettingsBox->Bounds();
   c_ButtonView.top += 15;
   c_ButtonView.left += 10;
   c_ButtonView.right -= 45;
   pc_TargetText = new BTextControl(c_ButtonView, "Target", NULL, NULL, NULL);
   pc_SettingsBox->AddChild(pc_TargetText);
   if(pc_LoopMessage->GetPath(&c_LoopPath) == B_OK)
      pc_TargetText->SetText(c_LoopPath.String());

   c_ButtonView = pc_SettingsBox->Bounds();
   c_ButtonView.top +=10;
   c_ButtonView.bottom = c_ButtonView.top + 10;
   c_ButtonView.right -= 10;
   c_ButtonView.left = c_ButtonView.right - 30;
   pc_PathButton = new BButton(c_ButtonView,"Path", "...",
         new BMessage(LOOP_SET_MSG_PATH));
   /* pc_PathButton->GetPreferredSize(&i_Width,&i_Height); */
   /* pc_PathButton->ResizeTo(i_Width,i_Height); */
   pc_SettingsBox->AddChild(pc_PathButton);

   c_ButtonView = pc_SettingsBox->Bounds();
   c_ButtonView.top += 45;
   c_ButtonView.left += 10;
   c_ButtonView.right = c_ButtonView.left + 25;
   pc_LoopNumberText = new BTextControl(c_ButtonView, "Loop Num", NULL, NULL, NULL);
   if(pc_LoopMessage->GetNumber(&i_Number) == B_OK)
   {
      char sz_String[16];

      snprintf(sz_String, 16, "%i", i_Number);
      sz_String[15] = '\0';
      
      pc_LoopNumberText->SetText(sz_String);
   }
   pc_SettingsBox->AddChild(pc_LoopNumberText);

   c_Point.x = c_ButtonView.left + 35;
   c_Point.y = c_ButtonView.top;
   pc_AlgoMenu = new AlgoMenu(c_Point);
   if(pc_LoopMessage->GetAlgo(&i_Algo) == B_OK)
   {
      pc_AlgoMenu->SetAlgo((CRYPT_ALGO_TYPE)i_Algo);
   }
   pc_SettingsBox->AddChild(pc_AlgoMenu->MenuField());
      
   /* keys box */
   c_ViewRect = pc_MainView->Bounds();
   c_ViewRect.left += 5;
   c_ViewRect.right -= 5;
   c_ViewRect.top += 100;
   c_ViewRect.bottom = c_ViewRect.top + 80;
   
   pc_KeyBox= new BBox(c_ViewRect);
   pc_KeyBox->SetLabel("Key Type");
   pc_MainView->AddChild(pc_KeyBox);
  
   /* ok/cancel buttons */
   c_ButtonView = Bounds();
   c_ButtonView.top += SETTINGS_WINDOW_HEIGHT - 30;
   c_ButtonView.left += 40;
   pc_OkButton = new BButton(c_ButtonView,"Apply", "Ok",
         new BMessage(LOOP_SET_MSG_OK));
   pc_OkButton->GetPreferredSize(&i_Width,&i_Height);
   pc_OkButton->ResizeTo(i_Width,i_Height);
   pc_MainView->AddChild(pc_OkButton);
   SetDefaultButton(pc_OkButton);

   c_ButtonView.left += 100;
   pc_OkButton = new BButton(c_ButtonView,"Cancel", "Cancel",
         new BMessage(LOOP_SET_MSG_CANCEL));
   pc_OkButton->GetPreferredSize(&i_Width,&i_Height);
   pc_OkButton->ResizeTo(i_Width,i_Height);
   pc_MainView->AddChild(pc_OkButton);


   /* set window position */
   BScreen  *pc_Screen = new BScreen();
   BRect c_WindowPosition(0, 0, 0+SETTINGS_WINDOW_WIDTH, 0+SETTINGS_WINDOW_HEIGHT);

   if(pc_Screen != NULL)
   {
      c_WindowPosition = pc_Screen->Frame();
      c_WindowPosition.left = c_WindowPosition.right - c_WindowPosition.left;
      c_WindowPosition.left /= 2;
      c_WindowPosition.left -= (SETTINGS_WINDOW_WIDTH/2);
      c_WindowPosition.right = c_WindowPosition.left + SETTINGS_WINDOW_WIDTH;

      c_WindowPosition.top = c_WindowPosition.bottom - c_WindowPosition.top;
      c_WindowPosition.top /= 2;
      c_WindowPosition.top -= (SETTINGS_WINDOW_HEIGHT/2);
      c_WindowPosition.bottom = c_WindowPosition.top + SETTINGS_WINDOW_HEIGHT;
   }
   
   MoveTo(c_WindowPosition.left, c_WindowPosition.top);
   
   i_OkSem = create_sem(0, "SettingsSem");

   Show();
   
}

SettingsWindow::~SettingsWindow(void)
{
   delete pc_LoopMessage;
}

void SettingsWindow::MessageReceived(BMessage *message)
{
   BString c_Path;
   
   switch (message->what)
   {
      case LOOP_SET_MSG_OK:
         pc_LoopMessage->SetNumber(atoi(pc_LoopNumberText->Text()));
         c_Path.SetTo(pc_TargetText->Text());
         pc_LoopMessage->SetPath(&c_Path);
         pc_LoopMessage->SetAlgo(pc_AlgoMenu->GetAlgo());

         if(b_NewLoop == true)
            pc_LoopPool->AddLoop(pc_LoopMessage);
         else
            pc_LoopPool->SetLoop(pc_LoopMessage);
         
         release_sem(i_OkSem);
         Quit();
         break;

      case LOOP_SET_MSG_CANCEL:
         b_Cancel = true;
         release_sem(i_OkSem);
         Quit();
         break;
      
      default:
         BWindow::MessageReceived(message);
         break;
   }
}

LoopMessage* SettingsWindow::Go(void)
{
   LoopMessage *pc_NewMessage;
   thread_id   i_Thread;
   BLooper  *pc_WindowLooper;
   BWindow  *pc_Window = NULL;
   
   i_Thread = find_thread(NULL);
   pc_WindowLooper = BLooper::LooperForThread(i_Thread);

   Show();
   
   if(pc_WindowLooper != NULL)
      pc_Window = cast_as(pc_WindowLooper, BWindow);
  
   /* if called from a window, poll to refresh it */
   if(pc_Window != NULL)
   {
      status_t i_ret;
      
      do
      {
         pc_Window->UpdateIfNeeded();
         i_ret = acquire_sem_etc(i_OkSem, 1, B_RELATIVE_TIMEOUT, 50000);
      } while( (i_ret == B_TIMED_OUT) || (i_ret == B_INTERRUPTED) );
   }
   else
   {   
      acquire_sem(i_OkSem);
   }

   /* acquire_sem(i_OkSem); */

   if(b_Cancel == false)
   {
      pc_NewMessage = new LoopMessage(*pc_LoopMessage);
      return(pc_NewMessage);
   }
   return(NULL);
}

bool SettingsWindow::QuitRequested()
{
   b_Cancel = true;
   release_sem(i_OkSem);
   return true;
}

