
/*
*****************************************************************************
*
*                          Copyright (c) 2000 2005
*
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice, 
*     this list of conditions and the following disclaimer.
*  - Redistributions in binary form must reproduce the above copyright notice, 
*     this list of conditions and the following disclaimer in the documentation
*     and/or other materials provided with the distribution.
*  - The names of its contributors may not be used to endorse or promote 
*     products derived from this software without specific prior written 
*     permission.
*   
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
*  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
*  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
*  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
*  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*  POSSIBILITY OF SUCH DAMAGE.
*
******************************************************************************
* File          : TextAlert.cpp
* Author        : Zadig
* Creation date : 27/08/2005
*.............................................................................
* Description :
******************************************************************************
*/

/* ########################################################################## */
/* ### #include                                                               */
#include <ClassInfo.h>
#include "TextAlert.h"

/* ########################################################################## */
/* ### Variables                                                              */

/* ########################################################################## */
/* ### Functions                                                              */

TextAlert::TextAlert(bool b_Password, char *sz_Title)  :
   BWindow(BRect(0, 0, TEXT_ALERT_WIDTH, TEXT_ALERT_HEIGHT), 
            sz_Title, B_TITLED_WINDOW_LOOK, 
            B_MODAL_APP_WINDOW_FEEL, 
            B_ASYNCHRONOUS_CONTROLS | B_NOT_RESIZABLE | B_NOT_ZOOMABLE )
{
   float i_Width, i_Height;
   BRect   c_ButtonView;
  
   pc_MainView = new BView(Bounds(), "MainView",          
         B_FOLLOW_LEFT,B_NAVIGABLE);
   pc_MainView->SetViewColor(TEXT_ALERT_BG_COLOR);
   AddChild(pc_MainView);

   c_ButtonView = Bounds();
   c_ButtonView.top += 10;
   c_ButtonView.left += 10;
   c_ButtonView.right -= 10;
   pc_TextView = new BTextControl(c_ButtonView, "Text", NULL, NULL, NULL);

   if(b_Password)
      (pc_TextView->TextView())->HideTyping(true);

   pc_MainView->AddChild(pc_TextView);
   pc_TextView->MakeFocus();
      
      
   c_ButtonView = Bounds();
   c_ButtonView.top += 40;
   c_ButtonView.left += 90;
   pc_OkButton = new BButton(c_ButtonView,"Apply", "Ok",
         new BMessage(TEXT_MSG_OK));
   pc_OkButton->GetPreferredSize(&i_Width,&i_Height);
   pc_OkButton->ResizeTo(i_Width,i_Height);
   pc_MainView->AddChild(pc_OkButton);

   SetDefaultButton(pc_OkButton);

   /* set window position */
   BScreen  *pc_Screen = new BScreen();
   BRect c_WindowPosition(0, 0, 0+TEXT_ALERT_WIDTH, 0+TEXT_ALERT_HEIGHT);

   if(pc_Screen != NULL)
   {
      c_WindowPosition = pc_Screen->Frame();
      c_WindowPosition.left = c_WindowPosition.right - c_WindowPosition.left;
      c_WindowPosition.left /= 2;
      c_WindowPosition.left -= (TEXT_ALERT_WIDTH/2);
      c_WindowPosition.right = c_WindowPosition.left + TEXT_ALERT_WIDTH;

      c_WindowPosition.top = c_WindowPosition.bottom - c_WindowPosition.top;
      c_WindowPosition.top /= 2;
      c_WindowPosition.top -= (TEXT_ALERT_HEIGHT/2);
      c_WindowPosition.bottom = c_WindowPosition.top + TEXT_ALERT_HEIGHT;
   }
   
   MoveTo(c_WindowPosition.left, c_WindowPosition.top);

   i_OkSem = create_sem(0, "TextSem");
   
}

TextAlert::~TextAlert(void)
{
   delete_sem(i_OkSem);
}

BString TextAlert::Go(void)
{
   thread_id   i_Thread;
   BLooper  *pc_WindowLooper;
   BWindow  *pc_Window = NULL;
   BString c_String;
   
   i_Thread = find_thread(NULL);
   pc_WindowLooper = BLooper::LooperForThread(i_Thread);
   
   Show();

   if(pc_WindowLooper != NULL)
      pc_Window = cast_as(pc_WindowLooper, BWindow);
  
   /* if called from a window, poll to refresh it */
   if(pc_Window != NULL)
   {
      status_t i_ret;
      
      do
      {
         pc_Window->UpdateIfNeeded();
         i_ret = acquire_sem_etc(i_OkSem, 1, B_RELATIVE_TIMEOUT, 50000);
      } while( (i_ret == B_TIMED_OUT) || (i_ret == B_INTERRUPTED) );
   }
   else
   {   
      acquire_sem(i_OkSem);
   }
   
   c_String.SetTo(pc_TextView->Text());

   /* close alert window */
   Lock();
   Quit();
   
   return(c_String);
}

void TextAlert::MessageReceived(BMessage *message)
{
   switch (message->what)
   {
      case TEXT_MSG_OK:
         release_sem(i_OkSem);
         break;

      default:
         BWindow::MessageReceived(message);
         break;
   }
}

