/*
*****************************************************************************
*
*                          Copyright (c) 2000 2005
*
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice, 
*     this list of conditions and the following disclaimer.
*  - Redistributions in binary form must reproduce the above copyright notice, 
*     this list of conditions and the following disclaimer in the documentation
*     and/or other materials provided with the distribution.
*  - The names of its contributors may not be used to endorse or promote 
*     products derived from this software without specific prior written 
*     permission.
*   
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
*  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
*  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
*  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
*  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*  POSSIBILITY OF SUCH DAMAGE.
*
******************************************************************************
* File          : desc.c
* Author        : Zadig
* Creation date : 27/01/2005
*.............................................................................
* Description :
******************************************************************************
*/

/* ########################################################################## */
/* ### #include                                                               */
#include <OS.h>
#include <malloc.h>

#include "loop.h"

/* ########################################################################## */
/* ### Functions                                                              */

/* ########################################################################## */
/* ### Variables                                                              */

TS_loop_filedesc * LOOP_NewDesc(void)
{
   TS_loop_filedesc * desc;
   
   desc = (TS_loop_filedesc *) malloc(sizeof(TS_loop_filedesc));
   if(desc == NULL)
      return NULL;
   desc->device = NULL;
   desc->sem_protect = create_sem(1, "Loop desc sem");
   if(desc->sem_protect < B_OK)
   {
      free(desc);
      return NULL;
   }
   
   if(lock_memory(&(desc->crypt), sizeof(TS_loop_crypt), 0) != B_OK)
   {
      delete_sem(desc->sem_protect);      
      free(desc);
      return NULL;
   }
   
   return(desc);
}

status_t LOOP_FreeDesc(TS_loop_filedesc *ps_Desc)
{
   if(ps_Desc == NULL)
      return EINVAL;
   
   ps_Desc->device  = NULL;

   /* paranoia = true; */
   /* change value several times to avoid bits remanence */
   memset(&ps_Desc->crypt, 0x55, sizeof(TS_loop_crypt));
   memset(&ps_Desc->crypt, 0xaa, sizeof(TS_loop_crypt));
   
   delete_sem(ps_Desc->sem_protect);
   /* NULLify everything to avoid pattern matching */
   memset(ps_Desc, 0, sizeof(TS_loop_filedesc));

   free(ps_Desc);
   return(B_OK);
}
