/*
*****************************************************************************
*
*                          Copyright (c) 2000 2005
*
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice, 
*     this list of conditions and the following disclaimer.
*  - Redistributions in binary form must reproduce the above copyright notice, 
*     this list of conditions and the following disclaimer in the documentation
*     and/or other materials provided with the distribution.
*  - The names of its contributors may not be used to endorse or promote 
*     products derived from this software without specific prior written 
*     permission.
*   
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
*  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
*  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
*  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
*  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*  POSSIBILITY OF SUCH DAMAGE.
*
******************************************************************************
* File          : device.c
* Author        : Zadig
* Creation date : 27/01/2005
*.............................................................................
* Description :
******************************************************************************
*/

/* ########################################################################## */
/* ### #include                                                               */
#include <unistd.h>
#include <drivers/driver_settings.h>

#include "loop.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* ########################################################################## */
/* ### Functions                                                              */

/* ########################################################################## */
/* ### Variables                                                              */

/****************************
 *    Device global init    *
 ****************************/
status_t LOOP_InitDevice(TS_loop_device *pas_Devices, uint8 i_NbDevice)
{
   uint8 i;

   sprintf(pas_Devices[0].name, "%s/loop_man", LOOP_BASE_PATH);
   pas_Devices[0].link[0] = '\0';
   pas_Devices[0].type = LOOP_MANAGER;
   pas_Devices[0].nb_open = 0;
   pas_Devices[0].sem_protect = create_sem(1, "Loop device sem");
   if(pas_Devices[0].sem_protect < B_OK)
      return(B_ERROR);

   for(i=1; i< i_NbDevice; i++)
   {
      pas_Devices[i].sem_protect = create_sem(1, "Loop device sem");      
      if(pas_Devices[i].sem_protect < B_OK)
         return(B_ERROR);

      /* prevent swaping on structure that contains crypt datas */
      if(lock_memory(&(pas_Devices[i]), sizeof(TS_loop_device), 0) != B_OK)
         return(B_ERROR);
         
      sprintf(pas_Devices[i].name, "%s/loop%d", LOOP_BASE_PATH, i-1);
      pas_Devices[i].link[0] = '\0';
      pas_Devices[i].type = LOOP_DEVICE;
      pas_Devices[i].nb_open = 0;
   }

   LOOP_ReadConfig(pas_Devices+1, i_NbDevice-1);
   return(B_OK);
}

status_t LOOP_FiniDevice(TS_loop_device *pas_Devices, uint8 i_NbDevice)
{
   uint8 i;

   for(i=0; i< i_NbDevice; i++)
   {
      delete_sem(pas_Devices[i].sem_protect);      
      unlock_memory(&(pas_Devices[i]), sizeof(TS_loop_device), 0);
   }

   return(B_OK);
}

TS_loop_device *LOOP_FindDevice(TS_loop_device *pas_Devices, 
                                 uint8 i_NbDevice,
                                 uint8 *sz_Device)
{
   uint8 i;

   if(strncmp(sz_Device, "/dev/", strlen("/dev/")) != 0)
      return(NULL);

   for(i=0; i< i_NbDevice; i++)
   {
      if(strcmp(pas_Devices[i].name, sz_Device + 5) == 0)
         return(&(pas_Devices[i]));
   }
   return(NULL);
}

status_t LOOP_ReadConfig(TS_loop_device *pas_Devices, uint8 i_NbDevice)
{
   void  *p_Handle;
   const driver_settings   *ps_Settings;

   p_Handle = load_driver_settings("loop");
   if(p_Handle != NULL)
   {
      ps_Settings = get_driver_settings(p_Handle);
      if(ps_Settings != NULL)
      {
         uint8 i_CurParam;
         driver_parameter  *ps_Parameter;
         
         /* get all loop configurations */
         for(i_CurParam=0; 
             i_CurParam < ps_Settings->parameter_count; 
             i_CurParam++)
         {
            ps_Parameter = &(ps_Settings->parameters[i_CurParam]);
            if(strcmp(ps_Parameter->name, "loop") == 0)
            {
               uint8 i_CurLoop;
               uint8             i_CurLoopParam;
               driver_parameter  *ps_LoopParameter;
               TS_loop_device    *ps_CurDevice;
               
               if(ps_Parameter->value_count == 1)
               {
                  i_CurLoop = atoi(ps_Parameter->values[0]);
                  ps_CurDevice = &(pas_Devices[i_CurLoop]);

                  dprintf("Loop Device : Found loop %d settings\n", i_CurLoop);
                  dprintf("Loop Device : loop %d param count = %d\n", i_CurLoop,ps_Parameter->parameter_count);
                  /* get all params for the current loop */
                  for(i_CurLoopParam=0; 
                      i_CurLoopParam< ps_Parameter->parameter_count; 
                      i_CurLoopParam++)
                  {
                     ps_LoopParameter = 
                        &(ps_Parameter->parameters[i_CurLoopParam]);
                     
                     if( (ps_LoopParameter->value_count == 1) &&
                         (strcmp(ps_LoopParameter->name, "link") == 0) )
                     {
                        dprintf("Loop Device : Found loop %d settings link\n", i_CurLoop);
                        strncpy(ps_CurDevice->link, 
                               ps_LoopParameter->values[0],
                               LOOP_STRING_SIZE);
                        ps_CurDevice->link[LOOP_STRING_SIZE-1] = '\0';
                     }
                  }
               }
            }            
         }
      }
      unload_driver_settings(p_Handle);
   }
   return(B_OK);
}


/*
******************************************************************************
*             Copyright (c) 2000 2006
*
* Function      : LOOP_SetDeviceGeometry
* Author        : Zadig
* Creation date : 17/09/2006
*.............................................................................
* Description   :
******************************************************************************
*/
status_t LOOP_SetDeviceGeometry(TS_loop_device *ps_Device, int f_File)
{
   struct stat st;
   status_t err = B_OK;
   
   if((fstat(f_File, &st) >= 0) && S_ISREG(st.st_mode) && (st.st_size > 0))
   {
      /* regular file: use stat() infos */
      /* make sure size is modulo block_size */
      st.st_size = st.st_size / ps_Device->block_size * ps_Device->block_size;
     
      /* compute old chs address style */
      ps_Device->geometry.bytes_per_sector = 512;
      ps_Device->geometry.sectors_per_track = 512;
      ps_Device->geometry.head_count = 8;
      ps_Device->geometry.cylinder_count = (uint32)(
            st.st_size / ps_Device->geometry.bytes_per_sector / 
            ps_Device->geometry.sectors_per_track / ps_Device->geometry.head_count);

      ps_Device->geometry.device_type = B_DISK;
      ps_Device->geometry.removable = false;
      ps_Device->geometry.read_only = false;
      ps_Device->geometry.write_once = false;
      ps_Device->dev_type = LOOP_DEVICE_FILE;
      DBG_PRINT("make_device_geometry from fstat");
   }
   else if(ioctl(f_File, B_GET_GEOMETRY, &ps_Device->geometry, 
                 sizeof(device_geometry)) >= 0)
   {
      /* block device: use existing geometry */
      ps_Device->dev_type = LOOP_DEVICE_DEV;
      DBG_PRINT("make_device_geometry from ioctl");
   }
   else
   {
      /* defaults */
      ps_Device->geometry.bytes_per_sector = 512;
      ps_Device->geometry.sectors_per_track = 0;
      ps_Device->geometry.cylinder_count = 0;
      ps_Device->geometry.head_count = 0;
      ps_Device->geometry.device_type = B_DISK;
      ps_Device->geometry.removable = false;
      ps_Device->geometry.read_only = false;
      ps_Device->geometry.write_once = false;
      ps_Device->dev_type = LOOP_DEVICE_FILE;
      DBG_PRINT("make_device_geometry from defaults");
   }
   
   return err;
}

