#ifndef _LOOP_H
#define _LOOPL_H

#include <OS.h>
#include <drivers/KernelExport.h>
#include <drivers/Drivers.h>

#include <sys/stat.h>
#include <cryptlib.h>
#include <CRYPT.h>
#include "cryptoloop.h"

/************************
 *       Defines        *
 ************************/
#define LOOP_DEVICE_MAX       8  /* nb of exported devices */
#define LOOP_DEVICE_NAME_SIZE LOOP_STRING_SIZE

#define LOOP_KEY_MIN       16 /* minimum key size */
#define LOOP_SALT_LENGHT   8
#if LOOP_KEY_MIN < LOOP_SALT_LENGHT
#error Min key size is too small!
#endif
#define LOOP_SALT_VALUE    "CrypLoop"

#define DBG_PRINT(s) dprintf("Loop Device: %s\n", s)

#define LOG_SIZE  (10*1024)

#ifndef IOCTL_FILE_UNCACHED_IO
#define IOCTL_FILE_UNCACHED_IO   10000
#endif

typedef enum
{
   LOOP_MANAGER,
   LOOP_DEVICE
} TE_loop_type;

typedef enum
{
   LOOP_DEVICE_FILE,
   LOOP_DEVICE_DEV
} TE_device_type;

/************************
 *       Structures     *
 ************************/
typedef struct
{
   int               iv_size;
   CRYPT_CONTEXT     context;
   CRYPT_CONTEXT     iv_context;
} TS_loop_crypt;

typedef struct
{
   sem_id         sem_protect;
   char           name[LOOP_DEVICE_NAME_SIZE];
   char           link[LOOP_STRING_SIZE];
   uint8          nb_open;
   TE_loop_type   type;
   TE_device_type dev_type;

   /* stat hook */
   struct stat    stat;
   ino_t          inode;

   /* crypt settings */
   uint8             passphrase[LOOP_STRING_SIZE];
   CRYPT_ALGO_TYPE   crypt_algo;
   uint32            block_size;
   device_geometry   geometry;
} TS_loop_device;

typedef struct
{
   sem_id         sem_protect;
   uint8          is_open;
   uint32         open_flags;
   int            file;
   TS_loop_device *device;
   
   TS_loop_crypt  crypt;
} TS_loop_filedesc;

/*************************
 *       Variables       *
 *************************/
extern TS_loop_device gas_loop_Devices[];
extern struct cryptlib_module_info  *gps_CryptMod;
/******************
 *    Functions   *
 ******************/

/* file crypt.c */
extern status_t LOOP_CryptInit(TS_loop_filedesc *ps_Desc);
extern status_t LOOP_CryptFini(TS_loop_filedesc *ps_Desc);
extern status_t LOOP_Crypt(TS_loop_filedesc *ps_Desc, uint32 i_Position, uint8 *p_Buf);
extern status_t LOOP_Decrypt(TS_loop_filedesc *ps_Desc, uint32 i_Position, uint8 *p_Buf);

/* file desc.c */
extern TS_loop_filedesc * LOOP_NewDesc(void);
extern status_t LOOP_FreeDesc(TS_loop_filedesc *ps_Desc);

/* file device.c */
extern status_t LOOP_InitDevice(TS_loop_device *pas_Devices, uint8 i_NbDevice);
extern status_t LOOP_FiniDevice(TS_loop_device *pas_Devices, uint8 i_NbDevice);
extern TS_loop_device *LOOP_FindDevice(TS_loop_device *pas_Devices,
                                       uint8 i_NbDevice,
                                       uint8 *sz_Device);
extern status_t LOOP_ReadConfig(TS_loop_device *pas_Devices, uint8 i_NbDevice);
extern status_t LOOP_SetDeviceGeometry(TS_loop_device *ps_Device, int f_File);

/* file icon.c */
extern status_t LOOP_GetIcon(device_icon *p_Data);

/* file manager.c */
extern status_t LOOP_ManSetLoop(TS_loop_device *ps_Device, 
                                TS_LOOP_loop_data *ps_Data);
extern status_t LOOP_ManGetLoop(TS_loop_device *ps_Device, 
                                TS_LOOP_loop_data *ps_Data);
extern status_t LOOP_ManSetAlgo(TS_loop_device *ps_Device, 
                                 TS_LOOP_algo_data *ps_Data);
extern status_t LOOP_ManGetAlgo(TS_loop_device *ps_Device, 
                                 TS_LOOP_algo_data *ps_Data);
extern status_t LOOP_ManSetKey(TS_loop_device *ps_Device, 
                                 TS_LOOP_key_data *ps_Data);
extern status_t LOOP_ManGetKey(TS_loop_device *ps_Device, 
                                 TS_LOOP_key_data *ps_Data);

/* file stat.c */
extern void LOOP_BuildInode(void);
extern status_t LOOP_StatHook(void);
extern status_t LOOP_StatRestore(void);


#endif
