/*
*****************************************************************************
*
*                          Copyright (c) 2000 2005
*
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice, 
*     this list of conditions and the following disclaimer.
*  - Redistributions in binary form must reproduce the above copyright notice, 
*     this list of conditions and the following disclaimer in the documentation
*     and/or other materials provided with the distribution.
*  - The names of its contributors may not be used to endorse or promote 
*     products derived from this software without specific prior written 
*     permission.
*   
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
*  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
*  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
*  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
*  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*  POSSIBILITY OF SUCH DAMAGE.
*
******************************************************************************
* File          : manager.c
* Author        : Zadig
* Creation date : 27/01/2005
*.............................................................................
* Description :
******************************************************************************
*/

/* ########################################################################## */
/* ### #include                                                               */
#include <string.h>

#include "loop.h"

/* ########################################################################## */
/* ### Functions                                                              */

/* ########################################################################## */
/* ### Variables                                                              */

/************************************************************
 *  Function : LOOP_ManSetLoop
 *  Date     : 30/01/2005
 ************************************************************
 * Usage:
 * Set the link pointed by a loop device. and initialise
 * default crypt settings.
 ************************************************************/
status_t LOOP_ManSetLoop(TS_loop_device *ps_Device, TS_LOOP_loop_data *ps_Data)
{
   acquire_sem(ps_Device->sem_protect);

   if(ps_Device->nb_open > 0)
   {
      release_sem(ps_Device->sem_protect);
      return(B_PERMISSION_DENIED);
   }   

   /* update link infos */
   strncpy(ps_Device->link, ps_Data->link, LOOP_STRING_SIZE);
   ps_Device->link[LOOP_STRING_SIZE-1] = '\0';
   
   if(ps_Device->link[0] == '\0')
   {
      /* if link is NULL, nothing to do (we deleted the link value) */
      release_sem(ps_Device->sem_protect);
      return(B_OK);
   }
   
   if(stat(ps_Device->link, &ps_Device->stat) != 0)
   {
      release_sem(ps_Device->sem_protect);
      return(B_PERMISSION_DENIED);
   }
   
   /* set the minimum block size. We can not set the fs block size here
      because the fs first reads a storage device block ie 512 bytes
      usually (this should not be hardcoded) */
   /* i_Size = ps_Device->stat.st_blksize; */
   ps_Device->block_size = 512;
   ps_Device->crypt_algo = CRYPT_ALGO_NONE;
   ps_Device->geometry.bytes_per_sector = 512;
   ps_Device->geometry.sectors_per_track = 0;
   ps_Device->geometry.cylinder_count = 0;
   ps_Device->geometry.head_count = 0;
   ps_Device->geometry.device_type = B_DISK;
   ps_Device->geometry.removable = false;
   ps_Device->geometry.read_only = false;
   ps_Device->geometry.write_once = false;

   release_sem(ps_Device->sem_protect);
   return(B_OK);
}

status_t LOOP_ManGetLoop(TS_loop_device *ps_Device, TS_LOOP_loop_data *ps_Data)
{
   acquire_sem(ps_Device->sem_protect);

   strncpy(ps_Data->link, ps_Device->link, LOOP_STRING_SIZE);
   ps_Data->link[LOOP_STRING_SIZE-1] = '\0';

   release_sem(ps_Device->sem_protect);
   return(B_OK);
}

status_t LOOP_ManSetAlgo(TS_loop_device *ps_Device, TS_LOOP_algo_data *ps_Data)
{
   acquire_sem(ps_Device->sem_protect);

   if(ps_Device->nb_open > 0)
   {
      release_sem(ps_Device->sem_protect);
      return(B_PERMISSION_DENIED);
   }

   /* update the algorithm used */
   ps_Device->crypt_algo = ps_Data->type;
   
   /* and clear the passphrase */
   memset(ps_Device->passphrase, 0, LOOP_STRING_SIZE);

   release_sem(ps_Device->sem_protect);
   return(B_OK);
}

status_t LOOP_ManGetAlgo(TS_loop_device *ps_Device, TS_LOOP_algo_data *ps_Data)
{
   acquire_sem(ps_Device->sem_protect);

   ps_Data->type = ps_Device->crypt_algo;

   release_sem(ps_Device->sem_protect);
   return(B_OK);
}

status_t LOOP_ManSetKey(TS_loop_device *ps_Device, TS_LOOP_key_data *ps_Data)
{
   acquire_sem(ps_Device->sem_protect);
   
   if(strlen(ps_Data->key) < LOOP_KEY_MIN)
   {
      release_sem(ps_Device->sem_protect);
      return(B_ERROR);
   }
   
   if(ps_Device->nb_open > 0)
   {
      release_sem(ps_Device->sem_protect);
      return(B_PERMISSION_DENIED);
   }   
   strncpy(ps_Device->passphrase, ps_Data->key, LOOP_STRING_SIZE);
   ps_Device->passphrase[LOOP_STRING_SIZE-1] = '\0';
   
   release_sem(ps_Device->sem_protect);
   return(B_OK);
}

status_t LOOP_ManGetKey(TS_loop_device *ps_Device, TS_LOOP_key_data *ps_Data)
{
#ifndef FLG_DBG
   return(B_PERMISSION_DENIED);
#else
   acquire_sem(ps_Device->sem_protect);

   strncpy(ps_Data->key, ps_Device->passphrase, LOOP_STRING_SIZE);
   ps_Data->key[LOOP_STRING_SIZE-1] = '\0';

   release_sem(ps_Device->sem_protect);
   return(B_OK);
#endif
}
