/*
*****************************************************************************
*
*                          Copyright (c) 2000 2005
*
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice, 
*     this list of conditions and the following disclaimer.
*  - Redistributions in binary form must reproduce the above copyright notice, 
*     this list of conditions and the following disclaimer in the documentation
*     and/or other materials provided with the distribution.
*  - The names of its contributors may not be used to endorse or promote 
*     products derived from this software without specific prior written 
*     permission.
*   
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
*  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
*  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
*  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
*  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
*  POSSIBILITY OF SUCH DAMAGE.
*
******************************************************************************
* File          : stat.c
* Author        : Zadig
* Creation date : 27/01/2005
*.............................................................................
* Description :
******************************************************************************
*/

/* ########################################################################## */
/* ### #include                                                               */
#include <stdio.h>
#include "loop.h"

/* ########################################################################## */
/* ### Functions                                                              */
extern int sys_rstat(void);
extern int sys_rstat_hook(struct stat *buf);

/* ########################################################################## */
/* ### Variables                                                              */
/* inode cache */
uint8 gb_loop_InodesDone = false;
uint8 gi_loop_Dev = 0xff;

/* hook infos */
#define START_SIZE   7
#define MAX_OFFSET   200

uint8 gas_HookStartData[] = {0x8b, 0x5d, 0xf8, 0x89, 0xec, 0x5d, 0xc3};
uint8 gai_StatStart[5];
uint32 gi_StartOffset;

void LOOP_BuildInode(void)
{
   uint8 i;
   char  sz_Name[LOOP_STRING_SIZE];
   TS_loop_device *ps_Device;
   
   /* build loop inode list */
   if(gb_loop_InodesDone == false)
   {
      for(i=0; i<LOOP_DEVICE_MAX; i++)
      {
         /* get loop file inode */
         ps_Device = &(gas_loop_Devices[i]);
         sprintf(sz_Name, "/dev/%s", ps_Device->name);
         if(stat(sz_Name, &ps_Device->stat) != 0)
            DBG_PRINT("loop stat failed\n");
         
         ps_Device->inode = ps_Device->stat.st_ino;
         if(gi_loop_Dev == 0xff)
            gi_loop_Dev = ps_Device->stat.st_dev;
         
         /* get link stats */
         if(ps_Device->link[0] != '\0')
         {
            if(stat(ps_Device->link, &ps_Device->stat) != 0)
               DBG_PRINT("loop link stat failed\n");
         }
      }
      gb_loop_InodesDone = true;
   }
}

int LOOP_SearchInode(ino_t i_Inode)
{
   uint8 i;

   /* search our inode */
   if(gb_loop_InodesDone == true)
   {
      for(i=0; i<LOOP_DEVICE_MAX; i++)
      {
         if(i_Inode == gas_loop_Devices[i].inode)
            return(i);
      }
   }
   return(-1);
}

void LOOP_Stat(struct stat *ps_Stat, uint32 b_FollowLink)
{
   int   i_Index;
      
   if(gb_loop_InodesDone == false)
      return;

   if(b_FollowLink == 0)
      return;
   
   if(ps_Stat->st_dev == gi_loop_Dev)
   {
      i_Index = LOOP_SearchInode(ps_Stat->st_ino);
      if( (i_Index >= 0) && (gas_loop_Devices[i_Index].link[0] != '\0') &&
          (gas_loop_Devices[i_Index].type == LOOP_DEVICE) )
      {
         /* we found the inode in the loop devices */
         /* update stat with the loop target values */
         *ps_Stat = gas_loop_Devices[i_Index].stat;
      }
   }
   return;
}

status_t LOOP_StatHook(void)
{
   uint8       *p_Stat, *p_MyStat;
   uint32      i_Offset;
   cpu_status  i_Status;

   p_Stat = (uint8*)sys_rstat;
   
   /* search the end of the function */
   gi_StartOffset = 0;
   while( (memcmp((p_Stat+ gi_StartOffset), gas_HookStartData, 
           START_SIZE) != 0) &&
          (gi_StartOffset < MAX_OFFSET) )
   {          
      gi_StartOffset++;
   }
   
   if(memcmp((p_Stat+gi_StartOffset), gas_HookStartData, START_SIZE) != 0)
   {
      DBG_PRINT("Error: Unable to hook sys_rstat()\n");
      return(B_ERROR);
   }
  
   /* apply the hook */
   p_Stat += gi_StartOffset;
   p_MyStat = (uint8*)sys_rstat_hook;
   i_Offset = (uint32)p_MyStat - ((uint32)p_Stat + 5);

   memcpy(gai_StatStart, p_Stat, 5);

   i_Status = disable_interrupts();
   *p_Stat = 0xE9; /* long jump */
   p_Stat++;
   *p_Stat = i_Offset & 0xFF;
   p_Stat++;
   *p_Stat = (i_Offset >> 8) & 0xFF;
   p_Stat++;
   *p_Stat = (i_Offset >> 16) & 0xFF;
   p_Stat++;
   *p_Stat = (i_Offset >> 24) & 0xFF;
   restore_interrupts(i_Status);

   DBG_PRINT("hooked sys_rstat()\n");
   return(B_OK);
}

status_t LOOP_StatRestore(void)
{
   uint8       *p_Stat;
   cpu_status  i_Status;
  
   if(gi_StartOffset > 0)
   {
      i_Status = disable_interrupts();
      p_Stat = (uint8*)sys_rstat;
      p_Stat += gi_StartOffset;
      memcpy(p_Stat, gai_StatStart, 5);
      restore_interrupts(i_Status);
      snooze(50000); /* wait 50ms to be sure that everybody is out of our hook */
      DBG_PRINT("restored sys_rstat()\n");
   }
   return(B_OK);
}

