#! /bin/sh

###########################################################
# make_files :
#param1 : base dir
###########################################################
make_files ()
{
   local sz_BaseDir="$1"
   
   dd if=/dev/random of="$sz_BaseDir/1" bs=1024k count=32 2>/dev/null
   dd if=/dev/random of="$sz_BaseDir/2" bs=1024k count=5 2>/dev/null
   dd if=/dev/random of="$sz_BaseDir/3" bs=1024k count=3 2>/dev/null
   dd if=/dev/random of="$sz_BaseDir/4" bs=1024k count=2 2>/dev/null
   dd if=/dev/random of="$sz_BaseDir/5" bs=1024k count=1 2>/dev/null
   dd if=/dev/random of="$sz_BaseDir/6" bs=1024k count=1 2>/dev/null
   dd if=/dev/random of="$sz_BaseDir/7" bs=512 count=1 2>/dev/null
   dd if=/dev/random of="$sz_BaseDir/8" bs=512 count=1 2>/dev/null
   dd if=/dev/random of="$sz_BaseDir/9" bs=32 count=5 2>/dev/null
   dd if=/dev/random of="$sz_BaseDir/10" bs=32 count=2 2>/dev/null
}

###########################################################
# make_part_files :
#param1 : base dir
###########################################################
make_part_files ()
{
   local sz_BaseDir="$1"
      
   dd if=/dev/zero of="$sz_BaseDir/1" bs=1024k count=110 2>/dev/null
   dd if=/dev/zero of="$sz_BaseDir/2" bs=1024k count=110 2>/dev/null
   dd if=/dev/zero of="$sz_BaseDir/3" bs=1024k count=110 2>/dev/null
   dd if=/dev/zero of="$sz_BaseDir/4" bs=1024k count=110 2>/dev/null
   dd if=/dev/zero of="$sz_BaseDir/5" bs=1024k count=110 2>/dev/null
   dd if=/dev/zero of="$sz_BaseDir/6" bs=1024k count=110 2>/dev/null
}

###########################################################
# do_setup :
#param1 : loop device
#param2 : loop target
#param3 : passphrase
###########################################################
do_setup ()
{
   (
    sleep 2
    echo "$3"
   ) | losetup -e aes "$1" "$2"
}

###########################################################
# mount_parts :
#param1 : base dir for the part files
###########################################################
mount_parts ()
{
   local sz_BaseDir="$1"
   local i=1
    
   while [ $i -le 6 ]
   do
      do_setup "/dev/disk/virtual/loop/loop$i" "$sz_BaseDir/$i" "ThetestPassPhraseNumber$i"
      mkbfs 2048 "/dev/disk/virtual/loop/loop$i" "test$i"  1>/dev/null
      sleep 5
      mkdir -p "/test$i"
      sleep 5
      mount -t bfs "/dev/disk/virtual/loop/loop$i" "/test$i"
      sleep 2
      mkdir "/test$i/files"
      i=$((i+1))
   done
}

###########################################################
# compare_tar :
#param1 : base dir for the tar file
###########################################################
compare_tar ()
{
   local sz_BaseDir="$1"
   local sz_Result=""
   local i=1
    
   while [ $i -le 6 ]
   do
      sz_Result=$(diff "$sz_BaseDir/test.tar" "/test$i/test.tar")
      if [ "$sz_Result" != "" ] ; then
         echo "Error while checking part $i"
      else
         echo "Part $i is ok"
      fi
      i=$((i+1))
   done
}

###########################################################
# compare_tar :
#param1 : base dir for the original files
#param2 : base dir for the copied files
###########################################################
compare_files ()
{
   local sz_OriDir="$1"
   local sz_DestDir="$2"
   local sz_Result=""
   local i=1
    
   while [ $i -le 10 ]
   do
      sz_Result=$(diff "$sz_OriDir/$i" "$sz_DestDir/$i")
      if [ "$sz_Result" != "" ] ; then
         echo "Error while checking file $2/$i"
      else
         echo "File $2/$i is ok"
      fi
      i=$((i+1))
   done
}

###########################################################
# touch_files :
#param1 : base dir for the touch
###########################################################
touch_files ()
{
   local i=1
   
   touch -t 01010101 "$1"
   
   while [ $i -le 10 ]
   do
      touch -t 01010101 "$1/$i"
      i=$((i+1))
   done
}

##################################################
# Script start
##################################################
# init
echo "****************************"
echo "*   CryptoLoop Test v1.1   *"
echo "*    by Zadig - 03/2005    *"
echo "****************************"

if [ "$1" == "" ] ; then
  echo "usage : test.sh [base dir]"
  exit
fi

sz_OldPath="$PWD"

echo "create all files..."
mkdir -p "$1/files"
mkdir -p "$1/parts"

#make_files "$1/files"
#make_part_files "$1/parts"

echo "mounting partitions..."
mount_parts "$1/parts"

echo "copying files..."
cp -r "$1/files/" /test1/ &
cp -r "$1/files/" /test2/ &
cp -r "$1/files/" /test3/ 
cp -r "$1/files/" /test4/ &
cp -r "$1/files/" /test5/ 
cp -r "$1/files/" /test6/
sleep 5

# touch all files so that they have the same date 
echo "touching files..."
touch_files "$1/files"
touch_files "/test1/files"
touch_files "/test2/files"
touch_files "/test3/files"
touch_files "/test4/files"
touch_files "/test5/files"
touch_files "/test6/files"
sleep 5

echo "comparing files..."
compare_files "$1/files" "/test1/files"
compare_files "$1/files" "/test2/files"
compare_files "$1/files" "/test3/files"
compare_files "$1/files" "/test4/files"
compare_files "$1/files" "/test5/files"
compare_files "$1/files" "/test6/files"

echo "cross taring files..."
cd "$1"
tar cvf "$1/test.tar" "files"

cd /test2
tar cvf /test1/test.tar files &
cd /test1
tar cvf /test2/test.tar files &
cd /test4
tar cvf /test3/test.tar files 
sleep 5

cd /test3
tar cvf /test4/test.tar files &
cd /test6
tar cvf /test5/test.tar files 
cd /test5
tar cvf /test6/test.tar files
sleep 5

echo "checking tar files..."
compare_tar "$1"
sync
sleep 10

echo "cleaning tmp files..."
unmount /test1
sleep 5
unmount /test2
sleep 5
unmount /test3
sleep 5
unmount /test4
sleep 5
unmount /test5
sleep 5
unmount /test6
sleep 5

rm -rf "$1/files"
rm -rf "$1/parts"
rm -f "$1/test.tar"

cd "$sz_OldPath"
